# -*- coding: utf-8 -*-

"""Communication(grpc) module."""

import grpc

from .grpc import account_pb2, blockchain_pb2, rpc_pb2, rpc_pb2_grpc
from .utils.converter import convert_tx_to_grpc_tx


class Comm:
    def __init__(self, target=None):
        self.__target = target
        self.__channel = None
        self.__rpc_stub = None

    def connect(self):
        self.disconnect()

        self.__channel = grpc.insecure_channel(self.__target)
        self.__rpc_stub = rpc_pb2_grpc.AergoRPCServiceStub(self.__channel)

    def disconnect(self):
        if self.__channel is not None:
            self.__channel.close()

    def create_account(self, address, passphrase):
        account = account_pb2.Account(address=address)
        return self.__rpc_stub.CreateAccount(request=rpc_pb2.Personal(account=account, passphrase=passphrase))

    def get_account_state(self, address):
        if self.__rpc_stub is None:
            return None

        rpc_account = account_pb2.Account()
        rpc_account.address = address
        return self.__rpc_stub.GetState(rpc_account)

    def get_account_state_proof(self, address, root, compressed):
        if self.__rpc_stub is None:
            return None
        account_and_root = rpc_pb2.AccountAndRoot(Account=address,
                                                  Root=root,
                                                  Compressed=compressed)
        return self.__rpc_stub.GetStateAndProof(account_and_root)

    def get_chain_info(self):
        if self.__rpc_stub is None:
            return None

        return self.__rpc_stub.GetChainInfo(rpc_pb2.Empty())

    def get_consensus_info(self):
        if self.__rpc_stub is None:
            return None

        return self.__rpc_stub.GetConsensusInfo(rpc_pb2.Empty())

    def get_blockchain_status(self):
        if self.__rpc_stub is None:
            return None

        return self.__rpc_stub.Blockchain(rpc_pb2.Empty())

    def get_accounts(self):
        if self.__rpc_stub is None:
            return None
        return self.__rpc_stub.GetAccounts(rpc_pb2.Empty())

    def get_block(self, block_hash):
        if self.__rpc_stub is None:
            return None

        v = rpc_pb2.SingleBytes()
        v.value = block_hash
        return self.__rpc_stub.GetBlock(v)

    def get_peers(self):
        return self.__rpc_stub.GetPeers(rpc_pb2.Empty())

    def get_node_state(self, timeout):
        single_bytes = rpc_pb2.SingleBytes()
        single_bytes.value = timeout.to_bytes(8, byteorder='little')
        return self.__rpc_stub.NodeState(single_bytes)

    def get_tx(self, tx_hash):
        single_bytes = rpc_pb2.SingleBytes()
        single_bytes.value = tx_hash
        return self.__rpc_stub.GetTX(single_bytes)

    def get_block_tx(self, tx_hash):
        single_bytes = rpc_pb2.SingleBytes()
        single_bytes.value = tx_hash
        return self.__rpc_stub.GetBlockTX(single_bytes)

    def unlock_account(self, address, passphrase):
        account = account_pb2.Account(address=address)
        personal = rpc_pb2.Personal(passphrase=passphrase, account=account)
        return self.__rpc_stub.UnlockAccount(request=personal)

    def lock_account(self, address, passphrase):
        account = account_pb2.Account(address=address)
        personal = rpc_pb2.Personal(passphrase=passphrase, account=account)
        return self.__rpc_stub.LockAccount(request=personal)

    # This RPC is for making and sending Tx inside a node.
    # Don't use it for sending TX which is made by a client.
    def send_tx(self, unsigned_tx):
        return self.__rpc_stub.SendTX(convert_tx_to_grpc_tx(unsigned_tx))

    # This RPC is for sending signed Txs made by a client.
    def commit_txs(self, signed_txs):
        rpc_txs = []
        for signed_tx in signed_txs:
            rpc_txs.append(convert_tx_to_grpc_tx(signed_tx))
        rpc_tx_list = blockchain_pb2.TxList()
        rpc_tx_list.txs.extend(rpc_txs)
        return self.__rpc_stub.CommitTX(rpc_tx_list)

    def get_receipt(self, tx_hash):
        if self.__rpc_stub is None:
            return None
        v = rpc_pb2.SingleBytes()
        v.value = tx_hash
        return self.__rpc_stub.GetReceipt(v)

    def query_contract(self, sc_address, query_info):
        query = blockchain_pb2.Query()
        query.contractAddress = sc_address
        query.queryinfo = query_info
        return self.__rpc_stub.QueryContract(query)

    def query_contract_state(self, sc_address, storage_keys, root, compressed):
        state_query = blockchain_pb2.StateQuery(contractAddress=sc_address,
                                                storageKeys=storage_keys,
                                                root=root,
                                                compressed=compressed)
        return self.__rpc_stub.QueryContractState(state_query)



import hashlib
import pytest

from aergo.herapy.obj.private_key import PrivateKey
from aergo.herapy.obj.address import Address
from aergo.herapy.utils.signature import verify_sig


def test_signature():
    msg = bytes("test", 'utf-8')
    h = hashlib.sha256(msg).digest()
    priv_key = PrivateKey(None)
    signature = priv_key.sign_msg(h)
    address = Address(priv_key.public_key).__str__()

    # check success
    assert verify_sig(h, signature, address)

    # check wrong signature from another key
    priv_key = PrivateKey(None)
    signature = priv_key.sign_msg(h)
    assert not verify_sig(h, signature, address)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import metric_pb2 as metric__pb2
from . import polarrpc_pb2 as polarrpc__pb2
from . import rpc_pb2 as rpc__pb2


class PolarisRPCServiceStub(object):
  # missing associated documentation comment in .proto file
  pass

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.NodeState = channel.unary_unary(
        '/types.PolarisRPCService/NodeState',
        request_serializer=rpc__pb2.NodeReq.SerializeToString,
        response_deserializer=rpc__pb2.SingleBytes.FromString,
        )
    self.Metric = channel.unary_unary(
        '/types.PolarisRPCService/Metric',
        request_serializer=metric__pb2.MetricsRequest.SerializeToString,
        response_deserializer=metric__pb2.Metrics.FromString,
        )
    self.CurrentList = channel.unary_unary(
        '/types.PolarisRPCService/CurrentList',
        request_serializer=polarrpc__pb2.Paginations.SerializeToString,
        response_deserializer=polarrpc__pb2.PolarisPeerList.FromString,
        )
    self.WhiteList = channel.unary_unary(
        '/types.PolarisRPCService/WhiteList',
        request_serializer=polarrpc__pb2.Paginations.SerializeToString,
        response_deserializer=polarrpc__pb2.PolarisPeerList.FromString,
        )
    self.BlackList = channel.unary_unary(
        '/types.PolarisRPCService/BlackList',
        request_serializer=polarrpc__pb2.Paginations.SerializeToString,
        response_deserializer=polarrpc__pb2.PolarisPeerList.FromString,
        )


class PolarisRPCServiceServicer(object):
  # missing associated documentation comment in .proto file
  pass

  def NodeState(self, request, context):
    """Returns the current state of this node
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Metric(self, request, context):
    """Returns node metrics according to request
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def CurrentList(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def WhiteList(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def BlackList(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_PolarisRPCServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'NodeState': grpc.unary_unary_rpc_method_handler(
          servicer.NodeState,
          request_deserializer=rpc__pb2.NodeReq.FromString,
          response_serializer=rpc__pb2.SingleBytes.SerializeToString,
      ),
      'Metric': grpc.unary_unary_rpc_method_handler(
          servicer.Metric,
          request_deserializer=metric__pb2.MetricsRequest.FromString,
          response_serializer=metric__pb2.Metrics.SerializeToString,
      ),
      'CurrentList': grpc.unary_unary_rpc_method_handler(
          servicer.CurrentList,
          request_deserializer=polarrpc__pb2.Paginations.FromString,
          response_serializer=polarrpc__pb2.PolarisPeerList.SerializeToString,
      ),
      'WhiteList': grpc.unary_unary_rpc_method_handler(
          servicer.WhiteList,
          request_deserializer=polarrpc__pb2.Paginations.FromString,
          response_serializer=polarrpc__pb2.PolarisPeerList.SerializeToString,
      ),
      'BlackList': grpc.unary_unary_rpc_method_handler(
          servicer.BlackList,
          request_deserializer=polarrpc__pb2.Paginations.FromString,
          response_serializer=polarrpc__pb2.PolarisPeerList.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'types.PolarisRPCService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# -*- coding: utf-8 -*-

import json

from .aer import Aer
from .chain_id import ChainID
from .consensus_info import ConsensusInfo


class BlockchainInfo:
    def __init__(self, chain_info, consensus_info=None):
        self._chain_info = chain_info
        self._chain_id = ChainID(self._chain_info.id)
        if consensus_info is not None:
            self._consensus_info = ConsensusInfo(consensus_info)
        else:
            self._consensus_info = ConsensusInfo(None,
                                                 self._chain_id.consensus)

    @property
    def number_of_bp(self):
        return self._chain_info.bpNumber

    @property
    def max_block_size(self):
        return self._chain_info.maxblocksize

    @property
    def max_tokens(self):
        return Aer(self._chain_info.maxtokens)

    @property
    def minimum_staking(self):
        return Aer(self._chain_info.stakingminimum)

    @property
    def total_staking(self):
        return Aer(self._chain_info.totalstaking)

    @property
    def gas_price(self):
        return Aer(self._chain_info.gasprice)

    @property
    def name_price(self):
        return Aer(self._chain_info.nameprice)

    @property
    def consensus_info(self):
        return self._consensus_info

    def json(self):
        return {
            "chain_id": self._chain_id.json(),
            "number_of_bp": str(self.number_of_bp),
            "max_block_size": self.max_block_size,
            "max_tokens": str(self.max_tokens),
            "minimum_staking": str(self.minimum_staking),
            "total_staking": str(self.total_staking),
            "gas_price": str(self.gas_price),
            "name_price": str(self.name_price),
            "consensus_info": self._consensus_info.json(),
        }

    def __str__(self):
        return json.dumps(self.json(), indent=2)


import sys
import traceback

import aergo.herapy as herapy


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)
    print(*args, **kwargs)


def run():
    try:
        aergo = herapy.Aergo()

        print("------ Connect AERGO -----------")
        aergo.connect('localhost:7845')

        print("------ Get Accounts without State in Node -----------")
        accounts = aergo.get_node_accounts(skip_state=True)
        for i, account in enumerate(accounts):
            print("  > account state : {}".format(account))
            print("    - balance")
            print("        + aergo              = {}".format(account.balance.aergo))
            print("        + gaer               = {}".format(account.balance.gaer))
            print("        + aer                = {}".format(account.balance.aer))
            print("    - nonce                  = {}".format(account.nonce))
            print("    - code hash              = {}".format(account.code_hash))
            print("    - storage root           = {}".format(account.storage_root))
            print("    - sql recovery point     = {}".format(account.sql_recovery_point))

        print("------ Get Accounts in Node -----------")
        accounts = aergo.get_node_accounts()
        for i, account in enumerate(accounts):
            print("  > account state : {}".format(account))
            print("    - balance")
            print("        + aergo              = {}".format(account.balance.aergo))
            print("        + gaer               = {}".format(account.balance.gaer))
            print("        + aer                = {}".format(account.balance.aer))
            print("    - nonce                  = {}".format(account.nonce))
            print("    - code hash              = {}".format(account.code_hash))
            print("    - storage root           = {}".format(account.storage_root))
            print("    - sql recovery point     = {}".format(account.sql_recovery_point))

        print("------ Create Account -----------")
        account = aergo.new_account()

        print("Private Key      = {}".format(account.private_key))
        print("Public Key       = {}".format(account.public_key))
        print("Address          = {}".format(account.address))

        print("------ Get Account State -----------")
        aergo.get_account()
        print("  > account state in 'aergo'")
        print("    - balance")
        print("        + aergo              = {}".format(account.balance.aergo))
        print("        + gaer               = {}".format(account.balance.gaer))
        print("        + aer                = {}".format(account.balance.aer))
        print("    - nonce                  = {}".format(account.nonce))
        print("    - code hash              = {}".format(account.code_hash))
        print("    - storage root           = {}".format(account.storage_root))
        print("    - sql recovery point     = {}".format(account.sql_recovery_point))
        print(account)

        print("------ Get Configured Account -----------")
        accounts = [
            {
                "private_key": "eHoEcHnaxpGpgzknXjuwon8VFVrLkKHC4FckGuGkQ8depiDDfyUAWC3L",
                "address": "AmPZKCJpT98V9Tc8dBUbRg78M1jgoB1ZEh97Rs1r5KewPcCiURf7",
            },
        ]
        for i, account in enumerate(accounts):
            print("  [{}]".format(i))
            print("    > private key   : {}".format(account['private_key']))
            print("    > address       : {}".format(account['address']))

            # check account state
            a = aergo.get_account(address=account['address'])
            print("    > account state : {}".format(a))
            print("    - balance")
            print("        + aergo              = {}".format(a.balance.aergo))
            print("        + gaer               = {}".format(a.balance.gaer))
            print("        + aer                = {}".format(a.balance.aer))
            print("    - nonce                  = {}".format(a.nonce))
            print("    - code hash              = {}".format(a.code_hash))
            print("    - storage root           = {}".format(a.storage_root))
            print("    - sql recovery point     = {}".format(a.sql_recovery_point))

        print("------ Disconnect AERGO -----------")
        aergo.disconnect()
    except Exception as e:
        eprint(e)
        traceback.print_exception(*sys.exc_info())


if __name__ == '__main__':
    run()



import sys
import traceback
import time

import aergo.herapy as herapy


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)
    print(*args, **kwargs)


def run():
    print("------ Payload -----------")
    """
        -- Define global variables.
        state.var {
            a = state.value(),
        }

        function constructor(name)
            a:set(name)
        end

        function set_name(name)
            a:set(name)
        end

        function get_name()
            return a:get()
        end

        function test_array(array)
            a:set(array[1])
        end

        function say_hello()
            a:set("hello")
        end


        abi.register(set_name, get_name, test_array, say_hello)
    """
    payload_str = "2bTAKwCcCqs1n91vFF5YnBBcaTLNsxhqzGgM8vcVAG3HqQLp1uZZSwvmnj8CEEm9GTvogDf3r6ioqHjbfBisjRzwUJxr6msFuf4vVuzbWEXtbShu43ZpLbJdT1NECStCEDGFPRsn3pc7jQ1FumPGB9GwGY7Y7wNN1t4kLMQPNxJCmDFAwLA6BAyCGpzW2g5rtHuWodTMTposF4EBRxbvsXYWvrDYLTK7JGbDsq4neBzXVtmJjereb7i6fX8Fug1LJ9GgYskTKdqcBjugxEoipjsAALtnr7mcUPoPFygbQzXbHsW13UwjhBxKsEvWYZm1hC8QxCeiRbHdx8VY5ES2ugR8Dt4uhiboAbaLofpnGRcPofgFGXiwDfESQgNZBjHP1DeEzh7gGh6fzABcw3LSjQLC9RETFWyUnWY3dU4iqH3PsVNhBFgHCUepBBphvwPT6UoictgsWzsnFdf1peiDsfXFE9uSTKWDCMhqgaETSLZdn8QiGBwMvZ1pEkP6xmh9389mFXCZ4ERNGUGkk6xNfN7xdZhRCeXxZupgtNDREX5PrHPveXEMbzThAMV7Rqzy3MazDaCFxjkbUVgQjWsHYhwEBRMvYSd2p621nZzcb9GFbh9tpXhwoLYpLp89Qhi9oTy61AD2GVFgQmdWF9neD1GvauVkCNFzWmEUkdSU3F6yBRwBm2cs8oJEieFu1zETLsMyXydTN5WqHFuGs2PaV1XRwPVRLeYiwo2xmgJMssHaGtR8pmb8TrQpTJbJdgreCVn8GiGzYizvEcRz9m8aFdompJ2m2QR7TDLCCjZx2UqRJdcQZmyxEAtzjnCCgD7Gwj95ZHkFxzoEEHTLYbG6uaneRrQWnzabAaRe2392qJnEkiNug9gkdJfsRFapbVxupH9zkoAmFP5xoWeGrkknUw61E4tX"
    payload = herapy.utils.decode_address(payload_str)
    print(''.join('{:d} '.format(x) for x in payload))

    byte_code_len = int.from_bytes(payload[:4], byteorder='little')
    print("payload: byte code length = ", byte_code_len)

    try:
        aergo = herapy.Aergo()

        print("------ Connect AERGO -----------")
        aergo.connect('localhost:7845')

        print("------ Set Sender Account -----------")
        sender_private_key = "6hbRWgddqcg2ZHE5NipM1xgwBDAKqLnCKhGvADWrWE18xAbX8sW"
        sender_account = aergo.new_account(private_key=sender_private_key)
        print("  > Sender Address: {}".format(sender_account.address))
        print(herapy.utils.convert_bytes_to_int_str(bytes(sender_account.address)))

        aergo.get_account()
        print("    > account state of Sender")
        print("      - balance        = {}".format(sender_account.balance))
        print("      - nonce          = {}".format(sender_account.nonce))
        print("      - code hash      = {}".format(sender_account.code_hash))
        print("      - storage root   = {}".format(sender_account.storage_root))

        print("------ Set Receiver Account -----------")
        receiver_address = "AmNHbk46L5ZaFH942mxDrunhUb34S8xRd7ygNnqaW5nqJDt5ugKD"
        print("  > Receiver Address: {}".format(receiver_address))

        print("------ Deploy SC -----------")
        tx, result = aergo.deploy_sc(amount=0, payload=payload, args=1234)
        print("  > TX: {}".format(tx.tx_hash))
        print("{}".format(herapy.utils.convert_tx_to_json(tx)))
        if result.status != herapy.CommitStatus.TX_OK:
            eprint("    > ERROR[{0}]: {1}".format(result.status, result.detail))
            aergo.disconnect()
            return
        else:
            print("    > result[{0}] : {1}".format(result.tx_id, result.status.name))
            print(herapy.utils.convert_bytes_to_int_str(bytes(tx.tx_hash)))

        time.sleep(3)
        aergo.get_account()

        print("------ Check deployment of SC -----------")
        print("  > TX: {}".format(tx.tx_hash))
        result = aergo.get_tx_result(tx.tx_hash)
        if result.status != herapy.TxResultStatus.CREATED:
            eprint("  > ERROR[{0}]:{1}: {2}".format(
                result.contract_address, result.status, result.detail))
            aergo.disconnect()
            return

        sc_address = result.contract_address
        print("  > SC Address: {}".format(sc_address))

        print("------ Fail SC -----------")
        tx, result = aergo.call_sc(sc_address, "set_none")
        time.sleep(3)
        result = aergo.get_tx_result(tx.tx_hash)
        print("  > ERROR[{0}]:{1}: {2}".format(
            result.contract_address, result.status, result.detail))

        print("------ Query SC -----------")
        result = aergo.query_sc(sc_address, "get_name")
        print(result)

        print("------ Call SC -----------")
        tx, result = aergo.call_sc(sc_address, "test_array", args=[["a", "b"]])

        time.sleep(3)

        print("------ Check result of Call SC -----------")
        print("  > TX: {}".format(tx.tx_hash))
        result = aergo.get_tx_result(tx.tx_hash)
        if result.status != herapy.TxResultStatus.SUCCESS:
            eprint("  > ERROR[{0}]:{1}: {2}".format(
                result.contract_address, result.status, result.detail))
            aergo.disconnect()
            return

        print("------ Query SC -----------")
        result = aergo.query_sc(sc_address, "get_name")
        print(result)

        print("------ Disconnect AERGO -----------")
        aergo.disconnect()
    except Exception as e:
        eprint(e)
        traceback.print_exception(*sys.exc_info())


if __name__ == '__main__':
    run()



import sys
import traceback
import time

import aergo.herapy as herapy
from aergo.herapy.obj.transaction import Transaction


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)
    print(*args, **kwargs)


def run():
    try:
        aergo = herapy.Aergo()

        print("------ Connect AERGO -----------")
        aergo.connect('localhost:7845')

        print("------ Set Sender Account -----------")
        sender_private_key = "eHoEcHnaxpGpgzknXjuwon8VFVrLkKHC4FckGuGkQ8depiDDfyUAWC3L"
        sender_account = aergo.new_account(private_key=sender_private_key)
        print("  > Sender Address: {}".format(sender_account.address))
        aergo.get_account()
        print("    > account state of Sender")
        print("      - balance        = {}".format(sender_account.balance))
        print("      - nonce          = {}".format(sender_account.nonce))
        print("      - code hash      = {}".format(sender_account.code_hash))
        print("      - storage root   = {}".format(sender_account.storage_root))

        print("------ Set Receiver Account -----------")
        receiver_address = "AmMPQqRJ4pd8bS9xdkw5pjExtLjaUXaYGB5kagzHu4A9ckKgnBV2"
        print("  > Receiver Address: {}".format(receiver_address))
        receiver_account = aergo.get_account(address=receiver_address)
        print("    > account state of Sender")
        print("      - balance        = {}".format(receiver_account.balance))
        print("      - nonce          = {}".format(receiver_account.nonce))
        print("      - code hash      = {}".format(receiver_account.code_hash))
        print("      - storage root   = {}".format(receiver_account.storage_root))

        print("------ Send batch Txs -----------")
        nonce = sender_account.nonce
        txs = [
            aergo.generate_tx(to_address=receiver_address, nonce=nonce+1, amount='9 aer'),
            aergo.generate_tx(to_address=receiver_address, nonce=nonce+2, amount='90 aer'),
            aergo.generate_tx(to_address=receiver_address, nonce=nonce+3, amount='900 aer'),
            aergo.generate_tx(to_address=receiver_address, nonce=nonce+4, amount='9000 aer'),
            aergo.generate_tx(to_address=receiver_address, nonce=nonce+5, amount='90000 aer'),
        ]

        txs, results = aergo.batch_tx(signed_txs=txs)
        for i, tx in enumerate(txs):
            print("[{}]".format(i))
            print("  > tx: ", str(tx))
            result = results[i]
            print("  > result: ", result)
            if result.status != herapy.CommitStatus.TX_OK:
                eprint("    > ERROR[{0}]: {1}".format(herapy.CommitStatus(result.status).name, result.detail))
            else:
                print("    > result[{0}] : {1}".format(result.tx_id, result.status.name))

        time.sleep(3)

        print("------ Check TX status -----------")
        for i, tx in enumerate(txs):
            print("[{}]".format(i))
            print("  > tx: ", str(tx))
            tx_result = aergo.get_tx_result(tx.tx_hash)
            print("      result: ", tx_result)
            if tx_result.status != herapy.TxResultStatus.SUCCESS:
                eprint("  > ERROR[{0}]:{1}: {2}".format(
                    tx_result.contract_address, tx_result.status, tx_result.detail))
                aergo.disconnect()
                return

        print("------ Get Sender Account Info -----------")
        aergo.get_account()
        print("    > account state of Sender")
        print("      - balance        = {}".format(sender_account.balance))
        print("      - nonce          = {}".format(sender_account.nonce))
        print("      - code hash      = {}".format(sender_account.code_hash))
        print("      - storage root   = {}".format(sender_account.storage_root))

        print("------ Disconnect AERGO -----------")
        aergo.disconnect()
    except Exception as e:
        eprint(e)
        traceback.print_exception(*sys.exc_info())


if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-

import json

from .address import Address
from .block_hash import BlockHash
from .transaction import Transaction
from ..utils.encoding import encode_b58
from ..utils.converter import get_hash


class Block:
    def __init__(self, hash_value=None, height=None, grpc_block=None):
        if grpc_block is not None:
            self._map_grpc_block(grpc_block)
            return

        if type(hash_value) is not BlockHash:
            hash_value = BlockHash(hash_value)

        self.__hash = hash_value
        # header
        self.__chain_id = None
        self.__prev_block = None
        self.__height = height
        self.__timestamp = None
        self.__blocks_root_hash = None
        self.__txs_root_hash = None
        self.__receipts_root_hash = None
        self.__confirms = None
        self.__public_key = None
        self.__sign = None
        self.__coinbase_account = None
        # body
        self.__tx_list = []

    def _map_grpc_block(self, v):
        self.__hash = BlockHash(v.hash)
        # header
        header = v.header
        self.__chain_id = header.chainID
        self.__prev_block = Block(hash_value=header.prevBlockHash,
                                  height=header.blockNo - 1)
        self.__height = header.blockNo
        self.__timestamp = header.timestamp
        self.__blocks_root_hash = header.blocksRootHash
        self.__txs_root_hash = header.txsRootHash
        self.__receipts_root_hash = header.receiptsRootHash
        self.__confirms = header.confirms
        self.__public_key = header.pubKey
        self.__sign = header.sign
        self.__coinbase_account = header.coinbaseAccount
        # body
        self.__tx_list = []
        for i, tx in enumerate(v.body.txs):
            from_address = Address(None, empty=True)
            from_address.value = tx.body.account
            to_address = Address(None, empty=True)
            to_address.value = tx.body.recipient

            block_tx = Transaction(read_only=True, tx_hash=tx.hash,
                                   nonce=tx.body.nonce,
                                   from_address=from_address,
                                   to_address=to_address,
                                   amount=tx.body.amount,
                                   payload=tx.body.payload,
                                   fee_price=tx.body.gasPrice,
                                   fee_limit=tx.body.gasLimit,
                                   tx_sign=tx.body.sign, tx_type=tx.body.type,
                                   block=self, index_in_block=i,
                                   is_in_mempool=False)
            self.__tx_list.append(block_tx)

    @property
    def hash(self):
        return self.__hash

    @property
    def chain_id(self):
        return self.__chain_id

    @property
    def chain_id_hash(self):
        if self.__chain_id is None:
            return None
        return get_hash(self.__chain_id, no_rand=True, no_encode=True)

    @property
    def chain_id_hash_b58(self):
        if self.__chain_id is None:
            return None
        return get_hash(self.__chain_id, no_rand=True, no_encode=False)

    @property
    def height(self):
        return self.__height

    @property
    def prev(self):
        return self.__prev_block

    @property
    def block_no(self):
        return self.height

    @property
    def timestamp(self):
        return self.__timestamp

    @property
    def blocks_root_hash(self):
        return self.__blocks_root_hash

    @property
    def txs_root_hash(self):
        return self.__txs_root_hash

    @property
    def receipts_root_hash(self):
        return self.__receipts_root_hash

    @property
    def confirms(self):
        return self.__confirms

    @property
    def public_key(self):
        return self.__public_key

    @property
    def sign(self):
        return self.__sign

    @property
    def coinbase_account(self):
        return self.__coinbase_account

    @property
    def tx_list(self):
        return self.__tx_list

    def get_tx(self, index):
        return self.__tx_list[index]

    def json(self, header_only=False):
        body_json = {
            "Hash": str(self.hash),
            "Header": {
                "ChainID": self.chain_id_hash_b58,
                "PreviousBlockHash": str(self.prev.hash) if self.prev is not None else None,
                "BlockNo": self.block_no,
                "Timestamp": self.timestamp,
                "BlocksRootHash": encode_b58(self.blocks_root_hash),
                "TxsRootHash": encode_b58(self.txs_root_hash),
                "ReceiptsRootHash": encode_b58(self.receipts_root_hash),
                "Confirms": self.confirms,
                "PubKey": encode_b58(self.public_key) if self.public_key is not None else None,
                "Sign": encode_b58(self.sign),
                "CoinbaseAccount": encode_b58(self.coinbase_account),
            },
        }

        if not header_only:
            tx_list = []
            for tx in self.tx_list:
                tx_list.append(tx.json())

            body_json["Body"] = {
                "Txs": tx_list,
            }

        return body_json

    def __str__(self):
        return json.dumps(self.json(), indent=2)



# -*- coding: utf-8 -*-

import ecdsa

from ..utils.encoding import encode_address, decode_address
from ..utils.converter import convert_bytes_to_public_key, \
    convert_public_key_to_bytes


class Address:
    def __init__(self, pubkey, empty=False, curve=ecdsa.SECP256k1):
        self.__address = None
        self.__curve = curve
        self.__empty = empty

        if empty:
            return

        if pubkey is None:
            assert 1 == 0

        if isinstance(pubkey, str):
            pubkey = decode_address(pubkey)
        elif isinstance(pubkey, bytes):
            pubkey = convert_bytes_to_public_key(pubkey, curve=curve)

        self.__address = convert_public_key_to_bytes(pubkey=pubkey,
                                                     curve=curve,
                                                     compressed=True)

    def __str__(self):
        return encode_address(self.__address)

    def __bytes__(self):
        return self.__address

    @property
    def value(self):
        return self.__address

    @value.setter
    def value(self, v):
        if self.__empty:
            if isinstance(v, str):
                self.__address = decode_address(v)
            elif isinstance(v, bytes):
                self.__address = v
        else:
            raise ValueError('Cannot set a value for the derived address')

    @property
    def curve(self):
        return self.__curve

    @property
    def public_key(self):
        return convert_bytes_to_public_key(self.__address, curve=self.__curve)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

from . import account_pb2 as account__pb2
from . import blockchain_pb2 as blockchain__pb2
from . import metric_pb2 as metric__pb2
from . import rpc_pb2 as rpc__pb2


class AergoRPCServiceStub(object):
  """*
  AergoRPCService is the main RPC service providing endpoints to interact 
  with the node and blockchain. If not otherwise noted, methods are unary requests.
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.NodeState = channel.unary_unary(
        '/types.AergoRPCService/NodeState',
        request_serializer=rpc__pb2.NodeReq.SerializeToString,
        response_deserializer=rpc__pb2.SingleBytes.FromString,
        )
    self.Metric = channel.unary_unary(
        '/types.AergoRPCService/Metric',
        request_serializer=metric__pb2.MetricsRequest.SerializeToString,
        response_deserializer=metric__pb2.Metrics.FromString,
        )
    self.Blockchain = channel.unary_unary(
        '/types.AergoRPCService/Blockchain',
        request_serializer=rpc__pb2.Empty.SerializeToString,
        response_deserializer=rpc__pb2.BlockchainStatus.FromString,
        )
    self.GetChainInfo = channel.unary_unary(
        '/types.AergoRPCService/GetChainInfo',
        request_serializer=rpc__pb2.Empty.SerializeToString,
        response_deserializer=rpc__pb2.ChainInfo.FromString,
        )
    self.ListBlockHeaders = channel.unary_unary(
        '/types.AergoRPCService/ListBlockHeaders',
        request_serializer=rpc__pb2.ListParams.SerializeToString,
        response_deserializer=rpc__pb2.BlockHeaderList.FromString,
        )
    self.ListBlockMetadata = channel.unary_unary(
        '/types.AergoRPCService/ListBlockMetadata',
        request_serializer=rpc__pb2.ListParams.SerializeToString,
        response_deserializer=rpc__pb2.BlockMetadataList.FromString,
        )
    self.ListBlockStream = channel.unary_stream(
        '/types.AergoRPCService/ListBlockStream',
        request_serializer=rpc__pb2.Empty.SerializeToString,
        response_deserializer=blockchain__pb2.Block.FromString,
        )
    self.ListBlockMetadataStream = channel.unary_stream(
        '/types.AergoRPCService/ListBlockMetadataStream',
        request_serializer=rpc__pb2.Empty.SerializeToString,
        response_deserializer=rpc__pb2.BlockMetadata.FromString,
        )
    self.GetBlock = channel.unary_unary(
        '/types.AergoRPCService/GetBlock',
        request_serializer=rpc__pb2.SingleBytes.SerializeToString,
        response_deserializer=blockchain__pb2.Block.FromString,
        )
    self.GetBlockMetadata = channel.unary_unary(
        '/types.AergoRPCService/GetBlockMetadata',
        request_serializer=rpc__pb2.SingleBytes.SerializeToString,
        response_deserializer=rpc__pb2.BlockMetadata.FromString,
        )
    self.GetBlockBody = channel.unary_unary(
        '/types.AergoRPCService/GetBlockBody',
        request_serializer=rpc__pb2.BlockBodyParams.SerializeToString,
        response_deserializer=rpc__pb2.BlockBodyPaged.FromString,
        )
    self.GetTX = channel.unary_unary(
        '/types.AergoRPCService/GetTX',
        request_serializer=rpc__pb2.SingleBytes.SerializeToString,
        response_deserializer=blockchain__pb2.Tx.FromString,
        )
    self.GetBlockTX = channel.unary_unary(
        '/types.AergoRPCService/GetBlockTX',
        request_serializer=rpc__pb2.SingleBytes.SerializeToString,
        response_deserializer=blockchain__pb2.TxInBlock.FromString,
        )
    self.GetReceipt = channel.unary_unary(
        '/types.AergoRPCService/GetReceipt',
        request_serializer=rpc__pb2.SingleBytes.SerializeToString,
        response_deserializer=blockchain__pb2.Receipt.FromString,
        )
    self.GetABI = channel.unary_unary(
        '/types.AergoRPCService/GetABI',
        request_serializer=rpc__pb2.SingleBytes.SerializeToString,
        response_deserializer=blockchain__pb2.ABI.FromString,
        )
    self.SendTX = channel.unary_unary(
        '/types.AergoRPCService/SendTX',
        request_serializer=blockchain__pb2.Tx.SerializeToString,
        response_deserializer=rpc__pb2.CommitResult.FromString,
        )
    self.SignTX = channel.unary_unary(
        '/types.AergoRPCService/SignTX',
        request_serializer=blockchain__pb2.Tx.SerializeToString,
        response_deserializer=blockchain__pb2.Tx.FromString,
        )
    self.VerifyTX = channel.unary_unary(
        '/types.AergoRPCService/VerifyTX',
        request_serializer=blockchain__pb2.Tx.SerializeToString,
        response_deserializer=rpc__pb2.VerifyResult.FromString,
        )
    self.CommitTX = channel.unary_unary(
        '/types.AergoRPCService/CommitTX',
        request_serializer=blockchain__pb2.TxList.SerializeToString,
        response_deserializer=rpc__pb2.CommitResultList.FromString,
        )
    self.GetState = channel.unary_unary(
        '/types.AergoRPCService/GetState',
        request_serializer=rpc__pb2.SingleBytes.SerializeToString,
        response_deserializer=blockchain__pb2.State.FromString,
        )
    self.GetStateAndProof = channel.unary_unary(
        '/types.AergoRPCService/GetStateAndProof',
        request_serializer=rpc__pb2.AccountAndRoot.SerializeToString,
        response_deserializer=blockchain__pb2.AccountProof.FromString,
        )
    self.CreateAccount = channel.unary_unary(
        '/types.AergoRPCService/CreateAccount',
        request_serializer=rpc__pb2.Personal.SerializeToString,
        response_deserializer=account__pb2.Account.FromString,
        )
    self.GetAccounts = channel.unary_unary(
        '/types.AergoRPCService/GetAccounts',
        request_serializer=rpc__pb2.Empty.SerializeToString,
        response_deserializer=account__pb2.AccountList.FromString,
        )
    self.LockAccount = channel.unary_unary(
        '/types.AergoRPCService/LockAccount',
        request_serializer=rpc__pb2.Personal.SerializeToString,
        response_deserializer=account__pb2.Account.FromString,
        )
    self.UnlockAccount = channel.unary_unary(
        '/types.AergoRPCService/UnlockAccount',
        request_serializer=rpc__pb2.Personal.SerializeToString,
        response_deserializer=account__pb2.Account.FromString,
        )
    self.ImportAccount = channel.unary_unary(
        '/types.AergoRPCService/ImportAccount',
        request_serializer=rpc__pb2.ImportFormat.SerializeToString,
        response_deserializer=account__pb2.Account.FromString,
        )
    self.ExportAccount = channel.unary_unary(
        '/types.AergoRPCService/ExportAccount',
        request_serializer=rpc__pb2.Personal.SerializeToString,
        response_deserializer=rpc__pb2.SingleBytes.FromString,
        )
    self.QueryContract = channel.unary_unary(
        '/types.AergoRPCService/QueryContract',
        request_serializer=blockchain__pb2.Query.SerializeToString,
        response_deserializer=rpc__pb2.SingleBytes.FromString,
        )
    self.QueryContractState = channel.unary_unary(
        '/types.AergoRPCService/QueryContractState',
        request_serializer=blockchain__pb2.StateQuery.SerializeToString,
        response_deserializer=blockchain__pb2.StateQueryProof.FromString,
        )
    self.GetPeers = channel.unary_unary(
        '/types.AergoRPCService/GetPeers',
        request_serializer=rpc__pb2.PeersParams.SerializeToString,
        response_deserializer=rpc__pb2.PeerList.FromString,
        )
    self.GetVotes = channel.unary_unary(
        '/types.AergoRPCService/GetVotes',
        request_serializer=rpc__pb2.VoteParams.SerializeToString,
        response_deserializer=rpc__pb2.VoteList.FromString,
        )
    self.GetAccountVotes = channel.unary_unary(
        '/types.AergoRPCService/GetAccountVotes',
        request_serializer=rpc__pb2.AccountAddress.SerializeToString,
        response_deserializer=rpc__pb2.AccountVoteInfo.FromString,
        )
    self.GetStaking = channel.unary_unary(
        '/types.AergoRPCService/GetStaking',
        request_serializer=rpc__pb2.AccountAddress.SerializeToString,
        response_deserializer=rpc__pb2.Staking.FromString,
        )
    self.GetNameInfo = channel.unary_unary(
        '/types.AergoRPCService/GetNameInfo',
        request_serializer=rpc__pb2.Name.SerializeToString,
        response_deserializer=rpc__pb2.NameInfo.FromString,
        )
    self.ListEventStream = channel.unary_stream(
        '/types.AergoRPCService/ListEventStream',
        request_serializer=blockchain__pb2.FilterInfo.SerializeToString,
        response_deserializer=blockchain__pb2.Event.FromString,
        )
    self.ListEvents = channel.unary_unary(
        '/types.AergoRPCService/ListEvents',
        request_serializer=blockchain__pb2.FilterInfo.SerializeToString,
        response_deserializer=rpc__pb2.EventList.FromString,
        )
    self.GetConsensusInfo = channel.unary_unary(
        '/types.AergoRPCService/GetConsensusInfo',
        request_serializer=rpc__pb2.Empty.SerializeToString,
        response_deserializer=rpc__pb2.ConsensusInfo.FromString,
        )


class AergoRPCServiceServicer(object):
  """*
  AergoRPCService is the main RPC service providing endpoints to interact 
  with the node and blockchain. If not otherwise noted, methods are unary requests.
  """

  def NodeState(self, request, context):
    """Returns the current state of this node
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Metric(self, request, context):
    """Returns node metrics according to request 
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Blockchain(self, request, context):
    """Returns current blockchain status (best block's height and hash)
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetChainInfo(self, request, context):
    """Returns current blockchain's basic information
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ListBlockHeaders(self, request, context):
    """Returns list of Blocks without body according to request
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ListBlockMetadata(self, request, context):
    """Returns list of block metadata (hash, header, and number of transactions) according to request
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ListBlockStream(self, request, context):
    """Returns a stream of new blocks as they get added to the blockchain
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ListBlockMetadataStream(self, request, context):
    """Returns a stream of new block's metadata as they get added to the blockchain
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetBlock(self, request, context):
    """Return a single block incl. header and body, queried by hash or number
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetBlockMetadata(self, request, context):
    """Return a single block's metdata (hash, header, and number of transactions), queried by hash or number
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetBlockBody(self, request, context):
    """Return a single block's body, queried by hash or number and list parameters
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTX(self, request, context):
    """Return a single transaction, queried by transaction hash
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetBlockTX(self, request, context):
    """Return information about transaction in block, queried by transaction hash
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetReceipt(self, request, context):
    """Return transaction receipt, queried by transaction hash
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetABI(self, request, context):
    """Return ABI stored at contract address
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SendTX(self, request, context):
    """Sign and send a transaction from an unlocked account
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SignTX(self, request, context):
    """Sign transaction with unlocked account
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def VerifyTX(self, request, context):
    """Verify validity of transaction
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def CommitTX(self, request, context):
    """Commit a signed transaction
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetState(self, request, context):
    """Return state of account
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetStateAndProof(self, request, context):
    """Return state of account, including merkle proof
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def CreateAccount(self, request, context):
    """Create a new account in this node
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAccounts(self, request, context):
    """Return list of accounts in this node
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def LockAccount(self, request, context):
    """Lock account in this node
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UnlockAccount(self, request, context):
    """Unlock account in this node
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ImportAccount(self, request, context):
    """Import account to this node
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ExportAccount(self, request, context):
    """Export account stored in this node
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def QueryContract(self, request, context):
    """Query a contract method
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def QueryContractState(self, request, context):
    """Query contract state
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetPeers(self, request, context):
    """Return list of peers of this node and their state
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetVotes(self, request, context):
    """Return result of vote
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetAccountVotes(self, request, context):
    """Return staking, voting info for account
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetStaking(self, request, context):
    """Return staking information
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetNameInfo(self, request, context):
    """Return name information
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ListEventStream(self, request, context):
    """Returns a stream of event as they get added to the blockchain
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def ListEvents(self, request, context):
    """Returns list of event
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetConsensusInfo(self, request, context):
    """Returns status of consensus and bps
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_AergoRPCServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'NodeState': grpc.unary_unary_rpc_method_handler(
          servicer.NodeState,
          request_deserializer=rpc__pb2.NodeReq.FromString,
          response_serializer=rpc__pb2.SingleBytes.SerializeToString,
      ),
      'Metric': grpc.unary_unary_rpc_method_handler(
          servicer.Metric,
          request_deserializer=metric__pb2.MetricsRequest.FromString,
          response_serializer=metric__pb2.Metrics.SerializeToString,
      ),
      'Blockchain': grpc.unary_unary_rpc_method_handler(
          servicer.Blockchain,
          request_deserializer=rpc__pb2.Empty.FromString,
          response_serializer=rpc__pb2.BlockchainStatus.SerializeToString,
      ),
      'GetChainInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetChainInfo,
          request_deserializer=rpc__pb2.Empty.FromString,
          response_serializer=rpc__pb2.ChainInfo.SerializeToString,
      ),
      'ListBlockHeaders': grpc.unary_unary_rpc_method_handler(
          servicer.ListBlockHeaders,
          request_deserializer=rpc__pb2.ListParams.FromString,
          response_serializer=rpc__pb2.BlockHeaderList.SerializeToString,
      ),
      'ListBlockMetadata': grpc.unary_unary_rpc_method_handler(
          servicer.ListBlockMetadata,
          request_deserializer=rpc__pb2.ListParams.FromString,
          response_serializer=rpc__pb2.BlockMetadataList.SerializeToString,
      ),
      'ListBlockStream': grpc.unary_stream_rpc_method_handler(
          servicer.ListBlockStream,
          request_deserializer=rpc__pb2.Empty.FromString,
          response_serializer=blockchain__pb2.Block.SerializeToString,
      ),
      'ListBlockMetadataStream': grpc.unary_stream_rpc_method_handler(
          servicer.ListBlockMetadataStream,
          request_deserializer=rpc__pb2.Empty.FromString,
          response_serializer=rpc__pb2.BlockMetadata.SerializeToString,
      ),
      'GetBlock': grpc.unary_unary_rpc_method_handler(
          servicer.GetBlock,
          request_deserializer=rpc__pb2.SingleBytes.FromString,
          response_serializer=blockchain__pb2.Block.SerializeToString,
      ),
      'GetBlockMetadata': grpc.unary_unary_rpc_method_handler(
          servicer.GetBlockMetadata,
          request_deserializer=rpc__pb2.SingleBytes.FromString,
          response_serializer=rpc__pb2.BlockMetadata.SerializeToString,
      ),
      'GetBlockBody': grpc.unary_unary_rpc_method_handler(
          servicer.GetBlockBody,
          request_deserializer=rpc__pb2.BlockBodyParams.FromString,
          response_serializer=rpc__pb2.BlockBodyPaged.SerializeToString,
      ),
      'GetTX': grpc.unary_unary_rpc_method_handler(
          servicer.GetTX,
          request_deserializer=rpc__pb2.SingleBytes.FromString,
          response_serializer=blockchain__pb2.Tx.SerializeToString,
      ),
      'GetBlockTX': grpc.unary_unary_rpc_method_handler(
          servicer.GetBlockTX,
          request_deserializer=rpc__pb2.SingleBytes.FromString,
          response_serializer=blockchain__pb2.TxInBlock.SerializeToString,
      ),
      'GetReceipt': grpc.unary_unary_rpc_method_handler(
          servicer.GetReceipt,
          request_deserializer=rpc__pb2.SingleBytes.FromString,
          response_serializer=blockchain__pb2.Receipt.SerializeToString,
      ),
      'GetABI': grpc.unary_unary_rpc_method_handler(
          servicer.GetABI,
          request_deserializer=rpc__pb2.SingleBytes.FromString,
          response_serializer=blockchain__pb2.ABI.SerializeToString,
      ),
      'SendTX': grpc.unary_unary_rpc_method_handler(
          servicer.SendTX,
          request_deserializer=blockchain__pb2.Tx.FromString,
          response_serializer=rpc__pb2.CommitResult.SerializeToString,
      ),
      'SignTX': grpc.unary_unary_rpc_method_handler(
          servicer.SignTX,
          request_deserializer=blockchain__pb2.Tx.FromString,
          response_serializer=blockchain__pb2.Tx.SerializeToString,
      ),
      'VerifyTX': grpc.unary_unary_rpc_method_handler(
          servicer.VerifyTX,
          request_deserializer=blockchain__pb2.Tx.FromString,
          response_serializer=rpc__pb2.VerifyResult.SerializeToString,
      ),
      'CommitTX': grpc.unary_unary_rpc_method_handler(
          servicer.CommitTX,
          request_deserializer=blockchain__pb2.TxList.FromString,
          response_serializer=rpc__pb2.CommitResultList.SerializeToString,
      ),
      'GetState': grpc.unary_unary_rpc_method_handler(
          servicer.GetState,
          request_deserializer=rpc__pb2.SingleBytes.FromString,
          response_serializer=blockchain__pb2.State.SerializeToString,
      ),
      'GetStateAndProof': grpc.unary_unary_rpc_method_handler(
          servicer.GetStateAndProof,
          request_deserializer=rpc__pb2.AccountAndRoot.FromString,
          response_serializer=blockchain__pb2.AccountProof.SerializeToString,
      ),
      'CreateAccount': grpc.unary_unary_rpc_method_handler(
          servicer.CreateAccount,
          request_deserializer=rpc__pb2.Personal.FromString,
          response_serializer=account__pb2.Account.SerializeToString,
      ),
      'GetAccounts': grpc.unary_unary_rpc_method_handler(
          servicer.GetAccounts,
          request_deserializer=rpc__pb2.Empty.FromString,
          response_serializer=account__pb2.AccountList.SerializeToString,
      ),
      'LockAccount': grpc.unary_unary_rpc_method_handler(
          servicer.LockAccount,
          request_deserializer=rpc__pb2.Personal.FromString,
          response_serializer=account__pb2.Account.SerializeToString,
      ),
      'UnlockAccount': grpc.unary_unary_rpc_method_handler(
          servicer.UnlockAccount,
          request_deserializer=rpc__pb2.Personal.FromString,
          response_serializer=account__pb2.Account.SerializeToString,
      ),
      'ImportAccount': grpc.unary_unary_rpc_method_handler(
          servicer.ImportAccount,
          request_deserializer=rpc__pb2.ImportFormat.FromString,
          response_serializer=account__pb2.Account.SerializeToString,
      ),
      'ExportAccount': grpc.unary_unary_rpc_method_handler(
          servicer.ExportAccount,
          request_deserializer=rpc__pb2.Personal.FromString,
          response_serializer=rpc__pb2.SingleBytes.SerializeToString,
      ),
      'QueryContract': grpc.unary_unary_rpc_method_handler(
          servicer.QueryContract,
          request_deserializer=blockchain__pb2.Query.FromString,
          response_serializer=rpc__pb2.SingleBytes.SerializeToString,
      ),
      'QueryContractState': grpc.unary_unary_rpc_method_handler(
          servicer.QueryContractState,
          request_deserializer=blockchain__pb2.StateQuery.FromString,
          response_serializer=blockchain__pb2.StateQueryProof.SerializeToString,
      ),
      'GetPeers': grpc.unary_unary_rpc_method_handler(
          servicer.GetPeers,
          request_deserializer=rpc__pb2.PeersParams.FromString,
          response_serializer=rpc__pb2.PeerList.SerializeToString,
      ),
      'GetVotes': grpc.unary_unary_rpc_method_handler(
          servicer.GetVotes,
          request_deserializer=rpc__pb2.VoteParams.FromString,
          response_serializer=rpc__pb2.VoteList.SerializeToString,
      ),
      'GetAccountVotes': grpc.unary_unary_rpc_method_handler(
          servicer.GetAccountVotes,
          request_deserializer=rpc__pb2.AccountAddress.FromString,
          response_serializer=rpc__pb2.AccountVoteInfo.SerializeToString,
      ),
      'GetStaking': grpc.unary_unary_rpc_method_handler(
          servicer.GetStaking,
          request_deserializer=rpc__pb2.AccountAddress.FromString,
          response_serializer=rpc__pb2.Staking.SerializeToString,
      ),
      'GetNameInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetNameInfo,
          request_deserializer=rpc__pb2.Name.FromString,
          response_serializer=rpc__pb2.NameInfo.SerializeToString,
      ),
      'ListEventStream': grpc.unary_stream_rpc_method_handler(
          servicer.ListEventStream,
          request_deserializer=blockchain__pb2.FilterInfo.FromString,
          response_serializer=blockchain__pb2.Event.SerializeToString,
      ),
      'ListEvents': grpc.unary_unary_rpc_method_handler(
          servicer.ListEvents,
          request_deserializer=blockchain__pb2.FilterInfo.FromString,
          response_serializer=rpc__pb2.EventList.SerializeToString,
      ),
      'GetConsensusInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetConsensusInfo,
          request_deserializer=rpc__pb2.Empty.FromString,
          response_serializer=rpc__pb2.ConsensusInfo.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'types.AergoRPCService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# -*- coding: utf-8 -*-

import ecdsa
import hashlib

from ecdsa.util import string_to_number

from . import address

from ..utils.encoding import encode_private_key, decode_private_key
from ..utils.signature import deserialize_sig, serialize_sig


class PrivateKey:
    def __init__(self, pk):
        if pk is not None:
            self.__get_key(pk)
        else:
            self.__generate_new_key()

        self.__address = address.Address(self.__private_key.public_key)

    def __str__(self):
        return encode_private_key(self.__get_private_key_bytes())

    def __bytes__(self):
        return self.__get_private_key_bytes()

    def __generate_new_key(self):
        self.__signing_key = ecdsa.SigningKey.generate(curve=ecdsa.SECP256k1,
                                                       hashfunc=hashlib.sha256)
        self.__private_key = self.__signing_key.privkey

    def __get_singing_key_instance(self, private_key):
        d = int.from_bytes(private_key, byteorder='big')
        """
        sk = ecdsa.SigningKey.from_secret_exponent(secexp=d,
                                                   curve=ecdsa.SECP256k1,
                                                   hashfunc=hashlib.sha256)
        """
        # to avoid 'Assert' error
        sk = ecdsa.SigningKey(_error__please_use_generate=True)
        sk.curve = ecdsa.SECP256k1
        sk.default_hashfunc = hashlib.sha256
        sk.baselen = ecdsa.SECP256k1.baselen
        pubkey_point = ecdsa.SECP256k1.generator * d
        pubkey = ecdsa.ecdsa.Public_key(ecdsa.SECP256k1.generator, pubkey_point)
        pubkey.order = ecdsa.SECP256k1.order
        sk.verifying_key = ecdsa.VerifyingKey.from_public_point(pubkey_point,
                                                                ecdsa.SECP256k1,
                                                                hashlib.sha256)
        sk.privkey = ecdsa.ecdsa.Private_key(pubkey, d)
        sk.privkey.order = ecdsa.SECP256k1.order
        return sk

    def __get_key(self, private_key):
        if isinstance(private_key, str):
            private_key = decode_private_key(private_key)

        self.__signing_key = self.__get_singing_key_instance(private_key)
        self.__private_key = self.__signing_key.privkey

    def __get_private_key_bytes(self):
        d = self.__private_key.secret_multiplier
        return d.to_bytes((d.bit_length() + 7) // 8, byteorder='big')

    def sign_msg(self, msg):
        r, s = self.__signing_key.sign_number(string_to_number(msg))
        order = self.__private_key.public_key.generator.order()
        return serialize_sig(r, s, order)

    def verify_sign(self, msg, sign):
        r, s = deserialize_sig(sign)
        signature = ecdsa.ecdsa.Signature(r, s)
        return self.__private_key.public_key.verifies(string_to_number(msg), signature)

    @property
    def public_key(self):
        return self.__private_key.public_key

    @property
    def address(self):
        return self.__address



# -*- coding: utf-8 -*-

"""Transaction class."""

import enum
import hashlib
import json

from . import tx_hash as txh
from ..obj import aer
from ..grpc import blockchain_pb2
from ..utils.encoding import encode_signature, decode_signature, encode_payload


@enum.unique
class TxType(enum.Enum):
    NORMAL = blockchain_pb2.NORMAL
    GOVERNANCE = blockchain_pb2.GOVERNANCE


class Transaction:
    """
    Transaction data structure:
    transaction = {
        hash : byte of base64
        nonce : int
        from : byte of base58
        to : byte of base58
        amount : uint
        payload : byte of base64
        type : int
        chain_id : bytes
        sign : byte of base64
    }
    """

    def __init__(self, from_address=None, to_address=None,
                 nonce=0, amount=0, payload=None, fee_price=0, fee_limit=0,
                 read_only=False, tx_hash=None, tx_sign=None, tx_type=TxType.NORMAL,
                 chain_id=None, block=None, index_in_block=-1, is_in_mempool=False):
        self.__from_address = from_address
        self.__to_address = to_address
        self.__nonce = nonce
        self.__amount = aer.Aer(amount)
        self.__payload = payload
        if isinstance(fee_price, bytes):
            fee_price = int.from_bytes(fee_price, byteorder='big')
        self.__fee_price = aer.Aer(fee_price)
        if isinstance(fee_limit, bytes):
            fee_limit = int.from_bytes(fee_limit, byteorder='little')
        self.__fee_limit = fee_limit
        if isinstance(tx_type, bytes):
            tx_type = int.from_bytes(tx_type, byteorder='little')
        self.__tx_type = TxType(tx_type)
        self.__read_only = read_only
        self.__tx_hash = txh.TxHash(tx_hash)
        if isinstance(tx_sign, str):
            tx_sign = decode_signature(tx_sign)
        self.__sign = tx_sign

        self.__chain_id = chain_id

        self.__is_in_mempool = is_in_mempool
        if is_in_mempool:
            self.__block = None
            self.__index_in_block = -1
        else:
            self.__block = block
            self.__index_in_block = index_in_block

    def calculate_hash(self, including_sign=True):
        m = hashlib.sha256()
        # nonce
        b = self.__nonce.to_bytes(8, byteorder='little')
        m.update(b)
        # from
        m.update(bytes(self.__from_address))
        # to
        if self.__to_address is not None:
            m.update(bytes(self.__to_address))
        # amount
        b = bytes(self.__amount)
        m.update(b)
        # payload
        if self.__payload is None:
            m.update(b'')
        else:
            m.update(self.__payload)
        # fee: limit
        b = self.__fee_limit.to_bytes(8, byteorder='little')
        m.update(b)
        # fee: price
        b = bytes(self.__fee_price)
        m.update(b)
        # type
        b = self.__tx_type.value.to_bytes(4, byteorder='little')
        m.update(b)
        # chainIdHash
        m.update(self.__chain_id)
        # sign
        if including_sign and self.__sign is not None:
            m.update(self.__sign)

        return m.digest()

    @property
    def block(self):
        return self.__block

    @property
    def index_in_block(self):
        return self.__index_in_block

    @property
    def is_in_mempool(self):
        return self.__is_in_mempool

    @property
    def nonce(self):
        return self.__nonce

    @nonce.setter
    def nonce(self, v):
        if self.__read_only:
            return

        self.__nonce = v

    @property
    def from_address(self):
        return self.__from_address

    @from_address.setter
    def from_address(self, v):
        if self.__read_only:
            return

        self.__from_address = v

    @property
    def to_address(self):
        return self.__to_address

    @to_address.setter
    def to_address(self, v):
        if self.__read_only:
            return

        self.__to_address = v

    @property
    def amount(self):
        return self.__amount

    @amount.setter
    def amount(self, v):
        if self.__read_only:
            return

        self.__amount = aer.Aer(v)

    @property
    def payload(self):
        return self.__payload

    @payload.setter
    def payload(self, v):
        if self.__read_only:
            return

        self.__payload = v

    @property
    def payload_str(self):
        if self.__payload is None:
            return None
        return encode_payload(self.__payload)

    @property
    def fee_limit(self):
        return self.__fee_limit

    @fee_limit.setter
    def fee_limit(self, v):
        if self.__read_only:
            return

        self.__fee_limit = v

    @property
    def fee_price(self):
        return self.__fee_price

    @fee_price.setter
    def fee_price(self, v):
        if self.__read_only:
            return

        self.__fee_price = v

    @property
    def tx_type(self):
        return self.__tx_type

    @tx_type.setter
    def tx_type(self, v):
        if self.__read_only:
            return

        self.__tx_type = TxType(v)

    @property
    def chain_id(self):
        return self.__chain_id

    @chain_id.setter
    def chain_id(self, v):
        if self.__read_only:
            return

        self.__chain_id = v

    @property
    def sign(self):
        return self.__sign

    @sign.setter
    def sign(self, v):
        if self.__read_only:
            return

        self.__sign = v

    @property
    def sign_str(self):
        if self.__sign is None:
            return None
        return encode_signature(self.__sign)

    @property
    def tx_hash(self):
        if self.__read_only:
            return self.__tx_hash
        return txh.TxHash(self.calculate_hash())

    def json(self, without_block=False):
        tx_json = {
            "Hash": str(self.tx_hash),
            "Body": {
                "Nonce": self.nonce,
                "Account": str(self.from_address) if self.from_address is not None else None,
                "Recipient": str(self.to_address) if self.to_address is not None else None,
                "Amount": str(self.amount),
                "Payload": self.payload_str,
                "Limit": self.fee_limit,
                "Price": str(self.fee_price),
                "Sign": self.sign_str,
                "Type": self.tx_type.name,
            },
            "IsInMempool": self.__is_in_mempool,
            "IndexInBlock": self.__index_in_block,
        }

        if not without_block:
            tx_json["Block"] = self.__block.json(header_only=True) if self.__block is not None else None,

        return tx_json

    def __str__(self):
        return json.dumps(self.json(), indent=2)

    def __bytes__(self):
        return self.tx_hash


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    "base58",
    "ecdsa",
    "googleapis-common-protos",
    "protobuf",
    "cryptography",
    "toml",
    "grpcio",
]

setup_requirements = [
    'setuptools',
    'flake8',
    'wheel',
    'twine',
    'Sphinx',
    'sphinx_rtd_theme',
]

test_requirements = [
    'grpcio-tools',
    'pytest',
    'tox',
    'codecov',
    'pytest-cov',
    'pytest-mock',
]

setup(
    name='aergo-herapy',
    version='1.0.3',
    description="python SDK for AERGO",
    keywords='herapy',
    author="aergo.io",
    author_email='tech@aergo.io',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
    ],
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    packages=find_packages(include=['aergo']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/aergoio/herapy',
    zip_safe=False,
)



# -*- coding: utf-8 -*-

import decimal

from ..errors.conversion_exception import ConversionException
from ..utils.converter import bigint_to_bytes

UNITS_SIZE = {
    'aergo': 18,
    'gaer': 9,
    'aer': 0
}
AERGO_UNIT_PRECISION = UNITS_SIZE['aergo']


class Aer:
    """Return Aergo Unit, AER(/ˈɛəɹ/)."""
    def __init__(self, value='0 aer'):
        if isinstance(value, bytes):
            value = int.from_bytes(value, byteorder='big')
        self.__aer = self._parsing_value(value)

    def _parsing_value(self, v):
        try:
            decimal.Decimal(v)
            v = self._decimal_to_str(v) + ' aer'
        except decimal.InvalidOperation:
            v = v.strip().lower()

        unit = ''
        size = -1
        for u, s in UNITS_SIZE.items():
            if v.endswith(u):
                unit = u
                size = s
                break

        if size < 0:
            raise ConversionException('cannot recognize aergo unit: ' + v)

        v = v[:len(v) - len(unit)].strip()
        if not '.' in v:
            zero_padding = UNITS_SIZE[unit]
        else:
            number, below = v.split('.')
            if len(below) > UNITS_SIZE[unit]:
                raise ConversionException('too small value below decimal point: ' + v)

            v = number + below
            zero_padding = UNITS_SIZE[unit] - len(below)

        v = v + '0' * zero_padding

        try:
            decimal.getcontext().prec = AERGO_UNIT_PRECISION
            v = decimal.Decimal(v)
        except decimal.InvalidOperation:
            raise ConversionException('cannot recognize value: ' + v)

        return v

    @staticmethod
    def _decimal_to_str(d):
        s = str(d).lower()
        if 'e' in s:
            digits, exp = s.split('e')

            exp = int(exp)
            if exp > 0 and '.' in digits:
                exp = exp - len(digits.split('.')[1]) + 1

            digits = digits.replace('.', '').replace('-', '')

            zero_padding = '0' * (abs(exp) - 1)  # minus 1 for decimal point in the sci notation
            sign = '-' if d < 0 else ''

            if exp > 0:
                s = '{}{}{}'.format(sign, digits, zero_padding)
            else:
                s = '{}0.{}{}'.format(sign, zero_padding, digits)
        return s

    @property
    def aer(self):
        return self._decimal_to_str(self.__aer) + ' aer'

    @property
    def gaer(self):
        with decimal.localcontext() as ctx:
            ctx.prec = len(str(self.__aer))
            v = self.__aer / decimal.Decimal(10 ** UNITS_SIZE['gaer'])
            s = self._decimal_to_str(v) + ' gaer'
        return s

    @property
    def aergo(self):
        with decimal.localcontext() as ctx:
            ctx.prec = len(str(self.__aer))
            v = self.__aer / decimal.Decimal(10 ** UNITS_SIZE['aergo'])
            s = self._decimal_to_str(v) + ' aergo'
        return s

    @property
    def dec(self):
        return self.__aer

    @dec.setter
    def dec(self, v):
        self.__aer = self._parsing_value(v)

    def __str__(self):
        return self.aer

    def __int__(self):
        return int(self.dec)

    def __bytes__(self):
        return bigint_to_bytes(int(self))

    def __add__(self, other):
        with decimal.localcontext():
            decimal.setcontext(decimal.DefaultContext)
            v = self.dec + other.dec
        return Aer(str(v))

    def __sub__(self, other):
        with decimal.localcontext():
            decimal.setcontext(decimal.DefaultContext)
            v = self.dec - other.dec
        return Aer(str(v))

    def __mul__(self, other):
        with decimal.localcontext() as ctx:
            ctx.prec = decimal.MAX_PREC
            v = self.dec * other.dec
        return Aer(str(v))

    def __floordiv__(self, other):
        with decimal.localcontext():
            decimal.setcontext(decimal.DefaultContext)
            v = self.dec // other.dec
        return Aer(str(v))

    def __truediv__(self, other):
        with decimal.localcontext():
            decimal.setcontext(decimal.DefaultContext)
            v = self.dec / other.dec
        return Aer(str(v))




# -*- coding: utf-8 -*-

"""Enumeration of Smart contract Status."""

import enum


@enum.unique
class PeerStatus(enum.Enum):
    """
    github.com/aergoio/aergo/types/peerstate.go
    """
    STARTING = 0
    HANDSHAKING = 1
    RUNNING = 2
    DOWN = 3
    STOPPED = 4



# -*- coding: utf-8 -*-

from google.protobuf.json_format import MessageToJson


class SCStateVar:
    """
    SCStateVar represents each variable of a calling smart contract.
    If the variable is the 'state.var' type,
    you can skip 'array_index' and 'map_key'.
    If the variable is the 'state.array' type,
    use 'array_index' with the index number.
    If the variable is the 'state.map' type,
    use 'map_key' with the key name of the map.
    """
    def __init__(self, var_name, array_index=None, map_key=None, empty=False):
        if empty:
            return

        self.var_name = var_name
        self.array_indx = array_index
        self.map_key = map_key

        self.get_storage_key()

    def get_storage_key(self):
        if self.array_indx is not None:
            self.storage_key = '_sv_{0}-{1}'.format(self.var_name, self.array_indx)
        elif self.map_key is not None:
            self.storage_key = '_sv_{0}-{1}'.format(self.var_name, self.map_key)
        else:
            self.storage_key = '_sv_{0}'.format(self.var_name)

    def __str__(self):
        return self.storage_key


class SCState:
    """ SCState holds the inclusion/exclusion proofs of a contract
    state in the global trie and of a variable's value in the
    contract trie.
    SCState is returned by aergo.query_sc_state() for easy merkle
    proof verification give a root.
    """
    def __init__(self, account, var_proofs):
        self.__account = account
        self.__var_proofs = var_proofs

    def __str__(self):
        account_str = MessageToJson(self.__account.state_proof)
        var_str = str(self.__var_proofs)
        return account_str + var_str

    @property
    def account(self):
        return self.__account

    @property
    def var_proofs(self):
        return self.__var_proofs

    @var_proofs.setter
    def var_proofs(self, v):
        self.__var_proofs = v

    def verify_proof(self, root):
        """ verify that the given inclusion and exclusion proofs are correct """
        if not self.__account.verify_proof(root):
            return False

        sc_root = self.__account.state_proof.state.storageRoot
        return self.__var_proofs.verify_proof(sc_root)




# -*- coding: utf-8 -*-

import inspect


class AergoException(Exception):
    # Exception types
    Comm = "Communication Exception"
    Conv = "Conversion Exception"
    General = "General Exception"

    def __init__(self, error, exception_type):
        self.error = error
        self.exception_type = exception_type

    def __str__(self):
        print_err = "{0} : {1}".format(self.exception_type, self.error)

        return print_err


class CommunicationException(AergoException):
    def __init__(self, error):
        if hasattr(error, 'code') \
                and (inspect.isfunction(error.code) or inspect.ismethod(error.code)):
            self.error_code = error.code()
        else:
            self.error_code = None

        if hasattr(error, 'details') \
                and (inspect.isfunction(error.details) or inspect.ismethod(error.details)):
            self.error_details = error.details()
        else:
            self.error_details = None

        self.exception_type = AergoException.Comm

    def __str__(self):
        if self.error_code is None:
            if self.error_details is None:
                print_err = "{0}: {1}".format(self.exception_type, "cannot recognize an error detail")
            else:
                print_err = "{0}: {1}".format(self.exception_type, self.error_details)
        else:
            print_err = "{0} ({1}): {2}".format(self.exception_type, self.error_code, self.error_details)

        return print_err


"""
Use this code when you add another module's exception.
You can see AergoException's print(__str__) format by using only this code.
If you want another format make __str__ like CommunicationException.

class TestException(AergoException):
    def __init__(self, error):
        super().__init__(error, EXCEPTION_TYPE)

"""



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pmap.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import node_pb2 as node__pb2
from . import p2p_pb2 as p2p__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='pmap.proto',
  package='types',
  syntax='proto3',
  serialized_options=_b('Z\036github.com/aergoio/aergo/types'),
  serialized_pb=_b('\n\npmap.proto\x12\x05types\x1a\nnode.proto\x1a\tp2p.proto\"X\n\x08MapQuery\x12\x1d\n\x06status\x18\x01 \x01(\x0b\x32\r.types.Status\x12\r\n\x05\x61\x64\x64Me\x18\x02 \x01(\x08\x12\x0c\n\x04size\x18\x03 \x01(\x05\x12\x10\n\x08\x65xcludes\x18\x04 \x03(\x0c\"j\n\x0bMapResponse\x12#\n\x06status\x18\x01 \x01(\x0e\x32\x13.types.ResultStatus\x12%\n\taddresses\x18\x02 \x03(\x0b\x32\x12.types.PeerAddress\x12\x0f\n\x07message\x18\x03 \x01(\tB Z\x1egithub.com/aergoio/aergo/typesb\x06proto3')
  ,
  dependencies=[node__pb2.DESCRIPTOR,p2p__pb2.DESCRIPTOR,])




_MAPQUERY = _descriptor.Descriptor(
  name='MapQuery',
  full_name='types.MapQuery',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='types.MapQuery.status', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='addMe', full_name='types.MapQuery.addMe', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='size', full_name='types.MapQuery.size', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='excludes', full_name='types.MapQuery.excludes', index=3,
      number=4, type=12, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=44,
  serialized_end=132,
)


_MAPRESPONSE = _descriptor.Descriptor(
  name='MapResponse',
  full_name='types.MapResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='status', full_name='types.MapResponse.status', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='addresses', full_name='types.MapResponse.addresses', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message', full_name='types.MapResponse.message', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=134,
  serialized_end=240,
)

_MAPQUERY.fields_by_name['status'].message_type = p2p__pb2._STATUS
_MAPRESPONSE.fields_by_name['status'].enum_type = p2p__pb2._RESULTSTATUS
_MAPRESPONSE.fields_by_name['addresses'].message_type = node__pb2._PEERADDRESS
DESCRIPTOR.message_types_by_name['MapQuery'] = _MAPQUERY
DESCRIPTOR.message_types_by_name['MapResponse'] = _MAPRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

MapQuery = _reflection.GeneratedProtocolMessageType('MapQuery', (_message.Message,), dict(
  DESCRIPTOR = _MAPQUERY,
  __module__ = 'pmap_pb2'
  # @@protoc_insertion_point(class_scope:types.MapQuery)
  ))
_sym_db.RegisterMessage(MapQuery)

MapResponse = _reflection.GeneratedProtocolMessageType('MapResponse', (_message.Message,), dict(
  DESCRIPTOR = _MAPRESPONSE,
  __module__ = 'pmap_pb2'
  # @@protoc_insertion_point(class_scope:types.MapResponse)
  ))
_sym_db.RegisterMessage(MapResponse)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



import ecdsa
from ecdsa.util import number_to_string, string_to_number

import hashlib

from .encoding import decode_address


def canonicalize_int(n, order):
    b = number_to_string(n, order)
    if (b[0] & 80) != 0:
        b = bytes([0]) + b
    return b


def serialize_sig(r, s, order):
    half_order = order >> 1
    if s > half_order:
        s = order - s

    rb = canonicalize_int(r, order)
    sb = canonicalize_int(s, order)

    length = 4 + len(rb) + len(sb)
    b = b'\x30' + bytes([length])
    b += b'\x02' + bytes([len(rb)]) + rb
    b += b'\x02' + bytes([len(sb)]) + sb
    return b


def deserialize_sig(sig):
    idx = 0
    if b'\x30'[0] != sig[idx]:
        # TODO error handling
        return None, None

    idx += 1

    length = len(sig) - 2
    if length != sig[idx]:
        # TODO error handling
        return None, None

    idx += 1

    # check R bytes
    if b'\x02'[0] != sig[idx]:
        # TODO error handling
        return None, None

    idx += 1
    r_len = sig[idx]
    idx += 1
    rb = sig[idx:idx+r_len]
    idx += r_len

    # check S bytes
    if b'\x02'[0] != sig[idx]:
        # TODO error handling
        return None, None

    idx += 1
    s_len = sig[idx]
    idx += 1
    sb = sig[idx:idx+s_len]

    return string_to_number(rb), string_to_number(sb)


def uncompress_key(compressed_key_hex):
    """
    base source : https://stackoverflow.com/questions/43629265/deriving-an-ecdsa-uncompressed-public-key-from-a-compressed-one?rq=1
    The code from bitcointalk sometimes produces a hex string uncompressed key of uneven length.
    """
    curve = ecdsa.SECP256k1

    prefix = compressed_key_hex[0:2]
    x_hex = compressed_key_hex[2:66]
    x = int(x_hex, 16)
    p = curve.curve.p()

    y_square = (pow(x, 3, p)  + 7) % p
    y_square_square_root = pow(y_square, (p+1)//4, p)
    if ((prefix == "02" and y_square_square_root & 1) or
        (prefix == "03" and not y_square_square_root & 1)):
        y = (-y_square_square_root) % p
    else:
        y = y_square_square_root

    computed_y_hex = format(y, '064x')
    computed_uncompressed_key = "04" + x_hex + computed_y_hex

    return computed_uncompressed_key


def verify_sig(msg, sig, address):
    """
    Verify that the signature 'sig' of the message 'msg' was made by 'address')
    """
    # format signature
    r, s = deserialize_sig(sig)
    signature = ecdsa.ecdsa.Signature(r, s)

    # format message
    number = string_to_number(msg)

    # get uncompressed pubkey from Aergo address
    pubkey_compressed = decode_address(address).hex()
    pubkey_uncompressed = uncompress_key(pubkey_compressed)
    pubkey = bytes.fromhex(pubkey_uncompressed)[1:]

    # verify signature matches pubkey
    vk = ecdsa.VerifyingKey.from_string(pubkey,
                                        curve=ecdsa.SECP256k1,
                                        hashfunc=hashlib.sha256)
    return vk.pubkey.verifies(number, signature)
